from mock import MagicMock
from scrunch.folders import Folder
from scrunch.mutable_dataset import MutableDataset
from pycrunch.shoji import Entity, Catalog

from .test_datasets import AttributeDict
from .mock_session import MockSession


def _getitem(var):
    variables = {
        'foo': AttributeDict(url='http://example.foo/'),
        'bar': AttributeDict(url='http://example.bar/'),
    }
    return variables[var]


def test_move_here_simple():
    """ basic test assuring `Folder.move_here` doesn't throw a basestring
    exception '"""
    entity_mock = MagicMock()
    root_mock = MagicMock()
    root_mock.dataset.__getitem__.side_effect = _getitem

    folder = Folder(entity_mock, root_mock, MagicMock())
    folder.move_here(['foo', 'bar'])
    entity_mock.patch.assert_called()


def test_unique_folders():
    session = MockSession()
    dataset_url = 'http://host/api/datasets/abc/'
    folders_url = 'http://host/api/datasets/abc/folders/'
    public_url = 'http://host/api/datasets/abc/folders/public/'
    hidden_url = 'http://host/api/datasets/abc/folders/hidden/'
    secure_url = 'http://host/api/datasets/abc/folders/secure/'
    dataset_resource = Entity(session, **{
        "element": "shoji:entity",
        "self": dataset_url,
        "body": {
            "name": "test_dataset_project"
        },
        "catalogs": {
            "folders": folders_url,
        }
    })
    dataset_resource.variables = MagicMock()
    dataset_resource.settings = MagicMock()
    folders_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": folders_url,
        "index": {},
        "body": {
            "name": "Root"
        },
        "catalogs": {
            "public": public_url,
            "hidden": hidden_url,
            "secure": secure_url,
        }
    })
    public_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": public_url,
        "index": {},
        "body": {
            "name": "Public"
        },
    })
    hidden_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": hidden_url,
        "index": {},
        "body": {
            "name": "Hidden"
        },
    })
    secure_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": secure_url,
        "index": {},
        "body": {
            "name": "Secure"
        },
    })
    session.add_fixture(folders_url, folders_resource)
    session.add_fixture(public_url, public_resource)
    session.add_fixture(hidden_url, hidden_resource)
    session.add_fixture(secure_url, secure_resource)
    dataset = MutableDataset(dataset_resource)

    assert dataset.folders.public.name == "Public"
    assert dataset.folders.hidden.name == "Hidden"
    assert dataset.folders.secure.name == "Secure"


def test_legacy_without_public():
    session = MockSession()
    dataset_url = 'http://host/api/datasets/abc/'
    folders_url = 'http://host/api/datasets/abc/folders/'
    dataset_resource = Entity(session, **{
        "element": "shoji:entity",
        "self": dataset_url,
        "body": {
            "name": "test_dataset_project"
        },
        "catalogs": {
            "folders": folders_url,
        }
    })
    dataset_resource.variables = MagicMock()
    dataset_resource.settings = MagicMock()

    # In this old format, there is no `public` catalog to follow. Scrunch
    # should use the folder_root as the .public and .root
    folders_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": folders_url,
        "index": {},
        "body": {
            "name": "Root"
        },
        "catalogs": {
        }
    })
    session.add_fixture(folders_url, folders_resource)
    dataset = MutableDataset(dataset_resource)

    assert dataset.folders.root.name == "Root"
    # Note how the public endpoint has Root because it uses the old API response
    assert dataset.folders.public.name == "Root"


def test_unique_folders_no_secure():
    session = MockSession()
    dataset_url = 'http://host/api/datasets/abc/'
    folders_url = 'http://host/api/datasets/abc/folders/'
    public_url = 'http://host/api/datasets/abc/folders/public/'
    hidden_url = 'http://host/api/datasets/abc/folders/hidden/'
    dataset_resource = Entity(session, **{
        "element": "shoji:entity",
        "self": dataset_url,
        "body": {
            "name": "test_dataset_project"
        },
        "catalogs": {
            "folders": folders_url,
        }
    })
    dataset_resource.variables = MagicMock()
    dataset_resource.settings = MagicMock()
    folders_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": folders_url,
        "index": {},
        "body": {
            "name": "Root"
        },
        "catalogs": {
            "public": public_url,
            "hidden": hidden_url,
            # Viewer users don't have the secure folder available
            # "secure": secure_url,
        }
    })
    public_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": hidden_url,
        "index": {},
        "body": {
            "name": "Public"
        },
    })
    hidden_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": hidden_url,
        "index": {},
        "body": {
            "name": "Hidden"
        },
    })
    session.add_fixture(folders_url, folders_resource)
    session.add_fixture(public_url, public_resource)
    session.add_fixture(hidden_url, hidden_resource)
    dataset = MutableDataset(dataset_resource)

    assert dataset.folders.public.name == "Public"
    assert dataset.folders.hidden.name == "Hidden"
    assert not hasattr(dataset.folders, "secure")


def test_unique_folders_no_hidden():
    session = MockSession()
    dataset_url = 'http://host/api/datasets/abc/'
    folders_url = 'http://host/api/datasets/abc/folders/'
    public_url = 'http://host/api/datasets/abc/folders/public/'
    dataset_resource = Entity(session, **{
        "element": "shoji:entity",
        "self": dataset_url,
        "body": {
            "name": "test_dataset_project"
        },
        "catalogs": {
            "folders": folders_url,
        }
    })
    dataset_resource.variables = MagicMock()
    dataset_resource.settings = MagicMock()
    folders_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": folders_url,
        "index": {},
        "body": {
            "name": "Root"
        },
        "catalogs": {
            # Standard exposed catalogs
            "public": public_url,
            "personal": "./personal/",
            "parents": "./parents/",
            # Viewer users don't have the secure folder available
            # "secure": secure_url,
            # Viewers also don't get the hidden folder exposed
            # "hidden": hidden_url,
        }
    })
    public_resource = Catalog(session, **{
        "element": "shoji:catalog",
        "self": public_url,
        "index": {},
        "body": {
            "name": "Public"
        },
    })
    session.add_fixture(folders_url, folders_resource)
    session.add_fixture(public_url, public_resource)
    dataset = MutableDataset(dataset_resource)

    assert dataset.folders.public.name == "Public"
    assert not hasattr(dataset.folders, "secure")
    assert not hasattr(dataset.folders, "hidden")



# coding: utf-8
import os
import pytest
from unittest import TestCase

from scrunch import connect
from scrunch.scripts import ScriptExecutionError
from scrunch.mutable_dataset import get_mutable_dataset


HOST = os.environ["SCRUNCH_HOST"]
username = os.environ["SCRUNCH_USER"]
password = os.environ["SCRUNCH_PASS"]

site = connect(username, password, HOST)
assert site is not None, "Unable to connect to %s" % HOST

as_entity = lambda b: {
    "element": "shoji:entity",
    "body": b
}


class TestScripts(TestCase):
    def _create_ds(self):
        ds = site.datasets.create(as_entity({"name": "test_script"})).refresh()
        variable = ds.variables.create(
            as_entity(
                {
                    "name": "pk",
                    "alias": "pk",
                    "type": "numeric",
                }
            )
        )
        return ds, variable

    def test_execute(self):
        ds, variable = self._create_ds()
        scrunch_dataset = get_mutable_dataset(ds.body.id, site, editor=True)
        script = """
        RENAME pk TO varA;

        CHANGE TITLE IN varA WITH "Variable A";
        """
        scrunch_dataset.scripts.execute(script)
        variable.refresh()
        assert variable.body["alias"] == "varA"
        assert variable.body["name"] == "Variable A"
        ds.delete()

    def test_handle_error(self):
        ds, variable = self._create_ds()
        scrunch_dataset = get_mutable_dataset(ds.body.id, site)
        script = """BAD-RENAME pk TO varA;"""  # Bad syntax script
        with pytest.raises(ScriptExecutionError) as err:
            scrunch_dataset.scripts.execute(script)

        variable.refresh()
        assert variable.body["alias"] == "pk"  # Unchanged
        resolutions = err.value.resolutions[0]
        assert resolutions["column"] == 1
        assert resolutions["line"] == 1
        assert resolutions["command"] == 1
        assert resolutions["message"] == "Invalid command: BAD"

        # Script big enough to trigger async validation
        async_script = ["""BAD-RENAME pk TO varA;"""] * 50000
        async_script = "\n".join(async_script)
        assert len(async_script) > 2 ** 20  # This is the threshold for async
        with pytest.raises(ScriptExecutionError) as err:
            scrunch_dataset.scripts.execute(async_script)
        assert len(err.value.resolutions) == 50000  # All lines raised error
        resolutions = err.value.resolutions[0]
        assert resolutions["column"] == 1
        assert resolutions["line"] == 1
        assert resolutions["command"] == 1
        assert resolutions["message"] == "Invalid command: BAD"
        ds.delete()

    def test_revert_script(self):
        ds, variable = self._create_ds()
        scrunch_dataset = get_mutable_dataset(ds.body.id, site)
        s1 = "RENAME pk TO varA;"
        s2 = 'CHANGE TITLE IN varA WITH "Variable A";'

        scrunch_dataset.scripts.execute(s1)
        scrunch_dataset.scripts.execute(s2)

        scrunch_dataset.scripts.revert_to(script_number=0)  # Reverted
        variable.refresh()
        assert variable.body["name"] == "pk"
        ds.delete()

    def test_fetch_all_and_collapse(self):
        raise self.skipTest("Collapse is 504ing in the server.")
        ds, variable = self._create_ds()
        scrunch_dataset = get_mutable_dataset(ds.body.id, site)
        s1 = "RENAME pk TO varA;"
        s2 = 'CHANGE TITLE IN varA WITH "Variable A";'

        scrunch_dataset.scripts.execute(s1)
        scrunch_dataset.scripts.execute(s2)

        r = scrunch_dataset.scripts.all()
        assert len(r) == 2
        assert r[0].body["body"] == s1
        assert r[1].body["body"] == s2

        scrunch_dataset.scripts.collapse()

        r = scrunch_dataset.scripts.all()
        assert len(r) == 1
        ds.delete()



from collections import OrderedDict

from six import PY2

from scrunch.helpers import ReadOnly


class Category(ReadOnly):
    _MUTABLE_ATTRIBUTES = {'name', 'numeric_value', 'missing', 'selected', 'date'}
    _IMMUTABLE_ATTRIBUTES = {'id'}
    _ENTITY_ATTRIBUTES = _MUTABLE_ATTRIBUTES | _IMMUTABLE_ATTRIBUTES
    _NULLABLE_ATTRIBUTES = {"date", "numeric_value", "selected"}

    def __init__(self, variable_resource, category):
        super(Category, self).__init__(variable_resource)
        self._category = category

    def __getattr__(self, item):
        if item in self._ENTITY_ATTRIBUTES:
            if item == 'selected':
                # Default is False; not always present
                return self._category.get('selected', False)
            try:
                return self._category[item]
            except KeyError:
                if item in self._NULLABLE_ATTRIBUTES:
                    return None
                else:
                    raise  # API returned an incomplete category? Error

        # Attribute doesn't exist, must raise an AttributeError
        raise AttributeError(
            'Category %s has no attribute %s' % (self.category['name'], item)
        )

    def __repr__(self):
        attrs = self.as_dict()
        return 'Category(%s)' % ', '.join('%s=%s' % c for c in attrs.items())

    def as_dict(self, **kwargs):
        attributes = self._ENTITY_ATTRIBUTES - {'date'}  # `date` needs special handling
        dct = {attr: getattr(self, attr) for attr in attributes}
        try:
            dct['date'] = getattr(self, 'date')
        except KeyError:
            # `date` is not there, just move on
            pass
        if "date" in dct and dct["date"] is None:
            del dct["date"]
        if PY2:
            dct['name'] = dct['name'].encode("ascii", "replace")
        dct.update(**kwargs or {})
        return dct

    def delete(self):
        if self.resource.body.get('derivation'):
            raise TypeError("Cannot delete categories on derived variables. Re-derive with the appropriate expression")

        categories = [cat for cat in self.resource.body['categories']
                      if cat['id'] != self.id]
        self.resource.edit(categories=categories)
        self.resource.refresh()

    def edit(self, **kwargs):
        if self.resource.body.get('derivation'):
            raise TypeError("Cannot edit categories on derived variables. Re-derive with the appropriate expression")
        extra_attrs = set(kwargs.keys()) - self._MUTABLE_ATTRIBUTES
        if extra_attrs:
            raise AttributeError("Cannot edit the following attributes: %s" % ', '.join(extra_attrs))

        categories = [self.as_dict(**kwargs) if cat['id'] == self.id else cat
                      for cat in self.resource.body['categories']]
        self.resource.edit(categories=categories)
        self.resource.refresh()


class CategoryList(OrderedDict):

    @classmethod
    def _from(cls, variable_resource):
        cls.resource = variable_resource
        categories = [(cat['id'], Category(variable_resource, cat))
                      for cat in variable_resource.body['categories']]
        return cls(categories)

    def order(self, *new_order):
        categories = sorted(
            self.resource.body['categories'], key=lambda c: new_order.index(c['id'])
        )
        self.resource.edit(categories=categories)
        self.resource.refresh()



#!/usr/bin/env python

import math
import os

import isodate
import pycrunch
import pytest

pytest.mark.skip('skip test discovery on this module')

try:
    from pycrunch import pandaslib
except ImportError:
    # pandas is not installed
    pandaslib = None

from scrunch import connect
from scrunch.datasets import Variable, get_geodata
from scrunch.streaming_dataset import StreamingDataset


CRUNCH_URL = os.environ.get('CRUNCH_TEST_URL')
CRUNCH_USER = os.environ.get('CRUNCH_TEST_USER')
CRUNCH_PASSWORD = os.environ.get('CRUNCH_TEST_PASSWORD')

# Metadata.
DATASET_DOC = {
    'body': {
        'name': 'scrunch test dataset',
        'description': 'scrunch integration tests',
        'table': {
            'element': 'crunch:table',
            'metadata': {
                'identity': {
                    'alias': 'identity',
                    'name': 'ID',
                    'type': 'numeric'
                },
                'ip_address': {
                    'alias': 'ip_address',
                    'name': 'Public IP Address',
                    'type': 'text'
                },
                'operating_system': {
                    'alias': 'operating_system',
                    'name': 'Operating System',
                    'type': 'text'
                },
                'registration_time': {
                    'alias': 'registration_time',
                    'name': 'Registration Time',
                    'resolution': 'ms',
                    'type': 'datetime'
                },
                'speak_spanish': {
                    'alias': 'speak_spanish',
                    'categories': [
                        {
                            'id': 1,
                            'missing': False,
                            'name': 'I speak Spanish primarily',
                            'numeric_value': 1
                        },
                        {
                            'id': 2,
                            'missing': False,
                            'name': 'I speak both Spanish and English equally',
                            'numeric_value': 2
                        },
                        {
                            'id': 3,
                            'missing': False,
                            'name': 'I speak English primarily but can speak Spanish',
                            'numeric_value': 3
                        },
                        {
                            'id': 4,
                            'missing': False,
                            'name': 'I can not speak Spanish',
                            'numeric_value': 4
                        },
                        {
                            'id': 32766,
                            'missing': True,
                            'name': 'skipped',
                            'numeric_value': None
                        },
                        {
                            'id': 32767,
                            'missing': True,
                            'name': 'not asked',
                            'numeric_value': None
                        },
                        {
                            'id': -1,
                            'missing': True,
                            'name': 'No Data',
                            'numeric_value': None
                        }
                    ],
                    'name': 'Do you speak Spanish?',
                    'type': 'categorical'
                },
                'hobbies': {
                    'alias': 'hobbies',
                    'categories': [
                        {
                            'id': 1,
                            'missing': False,
                            'name': 'Very interested',
                            'numeric_value': 1
                        },
                        {
                            'id': 2,
                            'missing': False,
                            'name': 'Somewhat interested',
                            'numeric_value': 2
                        },
                        {
                            'id': 3,
                            'missing': False,
                            'name': 'A little interested',
                            'numeric_value': 3
                        },
                        {
                            'id': 4,
                            'missing': False,
                            'name': 'Not at all interested',
                            'numeric_value': 4
                        },
                        {
                            'id': 32766,
                            'missing': True,
                            'name': 'skipped',
                            'numeric_value': None
                        },
                        {
                            'id': 32767,
                            'missing': True,
                            'name': 'not asked',
                            'numeric_value': None
                        },
                        {
                            'id': -1,
                            'missing': True,
                            'name': 'No Data',
                            'numeric_value': None
                        }
                    ],
                    'name': 'Hobbies',
                    'subvariables': [
                        {
                            'alias': 'hobbies_1',
                            'name': 'Sports'
                        },
                        {
                            'alias': 'hobbies_2',
                            'name': 'Video Games'
                        },
                        {
                            'alias': 'hobbies_3',
                            'name': 'Reading'
                        },
                        {
                            'alias': 'hobbies_4',
                            'name': 'Outdoor Activities'
                        }
                    ],
                    'type': 'categorical_array'
                },
                'music': {
                    'alias': 'music',
                    'categories': [
                        {
                            'id': 1,
                            'missing': False,
                            'name': 'selected',
                            'numeric_value': 1,
                            'selected': True
                        },
                        {
                            'id': 2,
                            'missing': False,
                            'name': 'not selected',
                            'numeric_value': 2,
                            'selected': False
                        },
                        {
                            'id': 32767,
                            'missing': True,
                            'name': 'not asked',
                            'numeric_value': None
                        },
                        {
                            'id': 32766,
                            'missing': True,
                            'name': 'skipped',
                            'numeric_value': None
                        },
                        {
                            'id': -1,
                            'missing': True,
                            'name': 'No Data',
                            'numeric_value': None
                        }
                    ],
                    'name': 'Music',
                    'subvariables': [
                        {
                            'alias': 'music_1',
                            'name': 'Pop'
                        },
                        {
                            'alias': 'music_2',
                            'name': 'Rock'
                        },
                        {
                            'alias': 'music_97',
                            'name': 'Other'
                        },
                        {
                            'alias': 'music_98',
                            'name': 'Don\'t know'
                        },
                        {
                            'alias': 'music_99',
                            'name': 'None of these'
                        }
                    ],
                    'type': 'multiple_response'
                },
                'religion': {
                    'alias': 'religion',
                    'categories': [
                        {
                            'id': 1,
                            'missing': False,
                            'name': 'Protestant',
                            'numeric_value': 1
                        },
                        {
                            'id': 2,
                            'missing': False,
                            'name': 'Catholic',
                            'numeric_value': 2
                        },
                        {
                            'id': 3,
                            'missing': False,
                            'name': 'Jewish',
                            'numeric_value': 3
                        },
                        {
                            'id': 4,
                            'missing': False,
                            'name': 'Muslim',
                            'numeric_value': 4
                        },
                        {
                            'id': 5,
                            'missing': False,
                            'name': 'None',
                            'numeric_value': 5
                        },
                        {
                            'id': 6,
                            'missing': False,
                            'name': 'Other',
                            'numeric_value': 6
                        },
                        {
                            'id': 32766,
                            'missing': True,
                            'name': 'skipped',
                            'numeric_value': None
                        },
                        {
                            'id': 32767,
                            'missing': True,
                            'name': 'not asked',
                            'numeric_value': None
                        },
                        {
                            'id': -1,
                            'missing': True,
                            'name': 'No Data',
                            'numeric_value': None
                        }
                    ],
                    'name': 'What is your religious preference?',
                    'type': 'categorical'
                },
                'location': {
                    'alias': 'location',
                    'categories': [
                        {
                            'id': 1,
                            'missing': False,
                            'name': 'Protestant',
                            'numeric_value': 1
                        }, {
                            'id': 2,
                            'missing': False,
                            'name': 'North East',
                            'numeric_value': 2
                        }, {
                            'id': 3,
                            'missing': False,
                            'name': 'North West',
                            'numeric_value': 3
                        }, {
                            'id': 4,
                            'missing': False,
                            'name': 'Yorkshire and the Humber',
                            'numeric_value': 4
                        }, {
                            'id': 5,
                            'missing': False,
                            'name': 'East Midlands',
                            'numeric_value': 5
                        }, {
                            'id': 6,
                            'missing': False,
                            'name': 'West Midlands',
                            'numeric_value': 6
                        }, {
                            'id': 7,
                            'missing': False,
                            'name': 'East of England',
                            'numeric_value': 7
                        }, {
                            'id': 8,
                            'missing': False,
                            'name': 'London',
                            'numeric_value': 8
                        }, {
                            'id': 9,
                            'missing': False,
                            'name': 'South East',
                            'numeric_value': 9
                        }, {
                            'id': 10,
                            'missing': False,
                            'name': 'South West',
                            'numeric_value': 10
                        }, {
                            'id': 11,
                            'missing': False,
                            'name': 'Wales',
                            'numeric_value': 11
                        }, {
                            'id': 12,
                            'missing': False,
                            'name': 'Scotland',
                            'numeric_value': 12
                        }, {
                            'id': 13,
                            'missing': False,
                            'name': 'Northern Ireland',
                            'numeric_value': 13
                        },
                    ],
                    'name': 'Where do you live?',
                    'type': 'categorical'
                }
            },
            'order': [
                {
                    'entities': [
                        'identity',
                        'ip_address',
                        'operating_system',
                        'registration_time',
                        'speak_spanish',
                        'hobbies',
                        'music',
                        'religion',
                        'location'
                    ],
                    'group': 'ungrouped'
                }
            ]
        }
    }
}

# Data
ROWS = [
    ['identity', 'ip_address', 'operating_system', 'registration_time', 'speak_spanish', 'hobbies_1', 'hobbies_2', 'hobbies_3', 'hobbies_4', 'music_1', 'music_2', 'music_97', 'music_98', 'music_99', 'religion', 'location'],
    [1, '10.0.0.1', 'Linux', '2014-04-21T10:00:00+00:00', 1, 32767, 32767, 32767, 32767, 2, 2, 1, 2, 2, 1, 1],
    [2, '10.0.0.2', 'Solaris', '2014-05-10T00:00:00+00:00', 1, 32766, 1, 1, 4, 1, 1, 1, 2, 2, 2, 2],
    [3, '10.0.0.3', 'Linux', '2015-01-01T00:00:00+00:00', 2, 2, 1, 2, 2, 2, 2, 2, 2, 1, 3, 3],
    [4, '10.0.0.4', 'Windows', '2015-01-02T00:00:00+00:00', 3, 4, 3, 2, 1, 1, 2, 1, 2, 2, 4, 4],
    [5, '10.0.0.5', 'Windows', '2015-02-01T00:00:00+00:00', 1, 1, 2, 32766, 4, 1, 1, 1, 2, 2, 5, 5],
    [6, '10.0.0.6', 'MacOS', '2015-06-01T00:00:00+00:00', 4, 2, 4, 4, 1, 2, 2, 1, 2, 2, 6, 6],
    [7, '10.0.0.7', 'Windows', '2015-12-30T00:00:00+00:00', 32766, 1, 32766, 4, 3, 2, 2, 2, 1, 2, 32766, 7],
    [8, '10.0.0.8', 'Minix', '2016-01-01T00:00:00+00:00', 32766, 2, 1, 1, 2, 2, 2, 2, 1, 2, 32767, 8],
    [9, '10.0.0.9', 'FreeBSD', '2016-02-01T00:00:00+00:00', 32767, 1, 1, 1, 32766, 1, 2, 1, 2, 2, 1, 9],
    [10, '10.0.0.10', 'NetBSD', '2015-03-01T00:00:00+00:00', 2, 4, 3, 4, 1, 2, 2, 1, 2, 2, 2, 10],
    [11, '10.0.0.10', 'NetBSD', '2015-03-01T00:01:00+00:00', 2, 4, 3, 4, 1, 1, 1, 1, 1, 1, 3, 11],
    [12, '10.0.0.10', 'NetBSD', '2015-03-01T00:02:00+00:00', 2, 4, 3, 4, 1, 2, 2, 2, 2, 2, 4, 12],
]


def invalid_credentials():
    return any(
        item is None
        for item in (CRUNCH_URL, CRUNCH_USER, CRUNCH_PASSWORD)
    )


def isnan(obj):
    if not isinstance(obj, float):
        return False
    return math.isnan(obj)


def main():
    assert not invalid_credentials()
    assert pandaslib, 'Pandas library not installed'

    # Login.
    site = connect(CRUNCH_USER, CRUNCH_PASSWORD, CRUNCH_URL)
    assert isinstance(site, pycrunch.shoji.Catalog)

    # Create the test dataset.
    dataset_resource = site.datasets.create(DATASET_DOC).refresh()
    assert isinstance(dataset_resource, pycrunch.shoji.Entity)
    dataset = StreamingDataset(dataset_resource)

    try:
        # Load initial data.
        pycrunch.importing.importer.append_rows(dataset.resource, ROWS)

        # refresh dataset instance, so size is updated
        dataset.resource.refresh()

        # Check the initial number of rows.
        df = pandaslib.dataframe(dataset.resource)
        assert len(df) == len(ROWS) - 1  # excluding the header
        assert dataset.size.rows == len(df)
        assert dataset.size.unfiltered_rows == len(df)

        # Also check number of columns
        columns = DATASET_DOC['body']['table']['metadata'].__len__()
        assert dataset.size.columns == columns

        # 0. Manipulate metadata

        # 0.1 Start by updating the missing rules for the `identity` variable
        identity_missing_rules = {
            "not asked": 9999,
            "skipped": 9998
        }

        assert dataset['identity'].missing_rules == {}
        dataset['identity'].set_missing_rules(identity_missing_rules)
        assert dataset['identity'].missing_rules == identity_missing_rules

        # 0.2 Try setting and unsetting the geodata view
        location = dataset['location']
        geodata = get_geodata('UK Regions')
        assert 'geodata' not in location.view

        # Set geodata using Entity object
        location.set_geodata_view(geodata, feature_key='EER13NM')
        assert 'geodata' in location.view

        location.unset_geodata_view()
        assert 'geodata' not in location.view

        # Set geodata using url
        location.set_geodata_view(geodata.self, feature_key='EER13NM')
        assert 'geodata' in location.view

        location.unset_geodata_view()
        assert 'geodata' not in location.view

        # Set geodata using name
        location.set_geodata_view('UK Regions', feature_key='EER13NM')
        assert 'geodata' in location.view

        location.unset_geodata_view()
        assert 'geodata' not in location.view

        # 1. Exclusion Filter Integration Tests

        # 1.1 Set a simple exclusion filter.

        dataset.exclude('identity > 5')
        df = pandaslib.dataframe(dataset.resource)
        assert len(df) == 5
        assert not any(r['identity'] > 5 for _, r in df.iterrows())

        # 1.2 More complex exclusion filters involving a categorical variable.

        expr = 'speak_spanish in [32766]'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] != 32766
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids

        expr = 'speak_spanish in (32766, 32767)'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] not in (32766, 32767)
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert not isnan(row['speak_spanish'])

        expr = 'not (speak_spanish in (1, 2) and operating_system == "Linux")'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] in (1, 2) and row[2] == 'Linux'
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['speak_spanish'] in \
                ('I speak Spanish primarily',
                    'I speak both Spanish and English equally')
            assert row['operating_system'] == 'Linux'

        # 1.3 Exclusion filters with `any`.

        expr = 'hobbies.any([32766])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and 32766 not in row[5:9]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert {'?': 32766} not in row['hobbies']

        expr = 'not hobbies.any([32766])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and 32766 in row[5:9]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert {'?': 32766} in row['hobbies']

        expr = 'hobbies.any([32766, 32767])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity'
            and 32766 not in row[5:9] and 32767 not in row[5:9]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert {'?': 32766} not in row['hobbies'] and \
                   {'?': 32767} not in row['hobbies']

        expr = 'music.any([32766])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and 32766 not in row[9:14]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert {'?': 32766} not in row['music']

        expr = 'music.any([1])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and 1 not in row[9:14]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert 1 not in row['music']

        expr = 'music.any([1, 2])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity'
            and 1 not in row[9:14] and 2 not in row[9:14]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert 1 not in row['music'] and 2 not in row['music']

        # 1.4 Exclusion filters with `all`.

        expr = 'hobbies.all([32767])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[5:9] != [32767, 32767, 32767, 32767]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['hobbies'] != [{'?': 32767}, {'?': 32767},
                                      {'?': 32767}, {'?': 32767}]

        expr = 'not hobbies.all([32767])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[5:9] == [32767, 32767, 32767, 32767]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['hobbies'] == [{'?': 32767}, {'?': 32767},
                                      {'?': 32767}, {'?': 32767}]

        expr = 'music.all([1])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[9:14] != [1, 1, 1, 1, 1]
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['music'] != [1, 1, 1, 1, 1]

        expr = 'music.all([1]) or music.all([2])'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity'
            and (row[9:14] != [1, 1, 1, 1, 1] and row[9:14] != [2, 2, 2, 2, 2])
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['music'] != [1, 1, 1, 1, 1] and \
                row['music'] != [2, 2, 2, 2, 2]

        expr = 'not ( music.all([1]) or music.all([2]) )'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity'
            and (row[9:14] == [1, 1, 1, 1, 1] or row[9:14] == [2, 2, 2, 2, 2])
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['music'] == [1, 1, 1, 1, 1] or \
                row['music'] == [2, 2, 2, 2, 2]

        # 1.5 Exclusion filters with `duplicates`.

        expr = 'ip_address.duplicates()'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        seen_ip_addresses = []
        for _, row in df.iterrows():
            assert row['ip_address'] not in seen_ip_addresses
            seen_ip_addresses.append(row['ip_address'])

        # 1.6 Exclusion filters with `valid` and `missing`.

        expr = 'valid(speak_spanish)'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] in (32766, 32767)
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert isnan(row['speak_spanish'])

        expr = 'not valid(speak_spanish)'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] not in (32766, 32767)
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert not isnan(row['speak_spanish'])

        expr = 'missing(speak_spanish)'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] not in (32766, 32767)
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert not isnan(row['speak_spanish'])

        expr = 'missing(hobbies)'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity'
            and (row[5:9] != [32766, 32766, 32766, 32766]
                 and row[5:9] != [32767, 32767, 32767, 32767])
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['hobbies'] != [{'?': 32766}, {'?': 32766},
                                      {'?': 32766}, {'?': 32766}] \
                and row['hobbies'] != [{'?': 32767}, {'?': 32767},
                                       {'?': 32767}, {'?': 32767}]

        expr = 'not missing(hobbies)'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity'
            and (row[5:9] == [32766, 32766, 32766, 32766]
                 or row[5:9] == [32767, 32767, 32767, 32767])
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['hobbies'] == [{'?': 32766}, {'?': 32766},
                                      {'?': 32766}, {'?': 32766}] \
                or row['hobbies'] == [{'?': 32767}, {'?': 32767},
                                      {'?': 32767}, {'?': 32767}]

        expr = 'valid(hobbies)'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and (32766 in row[5:9] or 32767 in row[5:9])
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert {'?': 32766} in row['hobbies'] or \
                   {'?': 32767} in row['hobbies']

        expr = 'not valid(hobbies)'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity'
            and (32766 not in row[5:9] and 32767 not in row[5:9])
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert {'?': 32766} not in row['hobbies'] and \
                   {'?': 32767} not in row['hobbies']

        # 1.7 Exclusion filter that refers to a subvariable by alias.
        expr = 'hobbies_1 == 4'
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and row[5] != 4
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids
            assert row['hobbies'][0] != 4

        # 1.8 Complex exclusion filters (multiple rules)
        expr = (
            '(religion != 1 and (not valid(speak_spanish) or speak_spanish >= 1)) '
            'or (religion == 1 and speak_spanish == 2) '
            'or (religion == 3 and speak_spanish == 4)'
        )
        dataset.exclude(expr)

        # 1.9 Exclusion filters using date variables.
        dt_str = '2014-12-30T00:00:00+00:00'
        dt = isodate.parse_datetime(dt_str)
        expr = 'registration_time < "%s"' % dt_str
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and isodate.parse_datetime(row[3]) >= dt
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids

        dt_str = '2015-01-01T00:00:00+00:00'
        dt = isodate.parse_datetime(dt_str)
        expr = 'registration_time >= "%s"' % dt_str
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and isodate.parse_datetime(row[3]) < dt
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids

        dt_str = '2014-05-10T00:00:00+00:00'
        dt = isodate.parse_datetime(dt_str)
        expr = 'registration_time == "%s"' % dt_str
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and isodate.parse_datetime(row[3]) != dt
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids

        dt_str = '2014-05-10T00:00:00+00:00'
        dt = isodate.parse_datetime(dt_str)
        expr = 'not(registration_time == "%s")' % dt_str
        dataset.exclude(expr)
        df = pandaslib.dataframe(dataset.resource)
        valid_ids = [
            row[0] for row in ROWS
            if row[0] != 'identity' and isodate.parse_datetime(row[3]) == dt
        ]
        assert len(df) == len(valid_ids)
        for _, row in df.iterrows():
            assert row['identity'] in valid_ids

        # 1.10 Clear the exclusion filter.
        dataset.exclude()
        df = pandaslib.dataframe(dataset.resource)
        assert len(df) == len(ROWS) - 1  # excluding the header

        # 2. Integration Tests for "Transformations".

        categories = [
            {
                'id': 1,
                'name': 'Nerds',
                'numeric_value': 1,
                'missing': False,
                'case': 'operating_system in ("Linux", "Solaris", "Minix", "FreeBSD", "NetBSD")',
            },
            {
                'id': 2,
                'name': 'Normal Users',
                'numeric_value': 2,
                'missing': False,
                'case': 'operating_system == "Windows"',
            },
            {
                'id': 3,
                'name': 'Hipsters',
                'numeric_value': 3,
                'missing': False,
                'case': 'operating_system == "MacOS"',
            },
            {
                'id': 32767,
                'name': 'Unknown',
                'numeric_value': None,
                'missing': True,
                'case': 'missing(operating_system)'
            }
        ]

        new_var = dataset.create_single_response(
            categories=categories,
            name='Operating System Users',
            alias='operating_system_users',
            description='Type of Operating System Users'
        )

        assert isinstance(new_var, Variable)
        assert new_var.type == 'categorical'

        # Check the data on the new variable.
        df = pandaslib.dataframe(dataset.resource)
        assert 'operating_system_users' in df

        # Check the nerds.
        assert set(
            r['operating_system']
            for _, r in df[df['operating_system_users'] == 'Nerds'].iterrows()
        ) == {'Linux', 'Solaris', 'Minix', 'FreeBSD', 'NetBSD'}

        # Check the hipsters.
        assert set(
            r['operating_system']
            for _, r in df[df['operating_system_users'] == 'Hipsters'].iterrows()
        ) == {'MacOS'}

        # Check normal users.
        assert set(
            r['operating_system']
            for _, r in df[df['operating_system_users'] == 'Normal Users'].iterrows()
        ) == {'Windows'}

        # 3. Integration Tests for "Recodes".

        # 3.1 combine_categories.

        # On a 'categorical' variable.
        cat_map = {
            1: [2, 3],
            2: [1, 4],
            99: [32766, 32767]
        }

        cat_names = {
            1: 'Bilingual',
            2: 'Not Bilingual',
            99: 'Unknown'
        }

        new_var = dataset.combine_categorical(
            'speak_spanish',
            map=cat_map,
            categories=cat_names,
            name='Bilingual Person',
            alias='bilingual',
            missing=[99]
        )

        assert isinstance(new_var, Variable)
        assert new_var.type == 'categorical'

        df = pandaslib.dataframe(dataset.resource)
        assert 'bilingual' in df

        # Check the data in the recoded variable.
        bilingual_ids = set(
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] in (2, 3)
        )
        assert set(
            int(r['identity'])
            for _, r in df[df['bilingual'] == 'Bilingual'].iterrows()
        ) == bilingual_ids

        non_bilingual_ids = set(
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] in (1, 4)
        )
        assert set(
            int(r['identity'])
            for _, r in df[df['bilingual'] == 'Not Bilingual'].iterrows()
        ) == non_bilingual_ids

        bilingual_null_ids = set(
            row[0] for row in ROWS
            if row[0] != 'identity' and row[4] in (32766, 32767)
        )
        assert set(
            int(r['identity'])
            for _, r in df[df['bilingual'].isnull()].iterrows()
        ) == bilingual_null_ids

        # On a 'categorical_array' variable.

        cat_map = {
            1: [1, 2],
            2: [3, 4],
            99: [32766, 32767]
        }

        cat_names = {
            1: 'Interested',
            2: 'Not interested',
            99: 'Unknown',
        }

        new_var = dataset.combine_categorical(
            'hobbies',
            map=cat_map,
            categories=cat_names,
            name='Hobbies (recoded)',
            alias='hobbies_recoded',
            missing=[99]
        )
        assert isinstance(new_var, Variable)
        assert new_var.type == 'categorical_array'

        df = pandaslib.dataframe(dataset.resource)
        assert 'hobbies_recoded' in df

        # Check the data in the recoded variable.
        for _, row in df.iterrows():
            hobbies = row['hobbies']
            hobbies_rec = row['hobbies_recoded']
            assert len(hobbies) == len(hobbies_rec)

            for i, value in enumerate(hobbies):
                if value in ({'?': 32766}, {'?': 32767}):
                    assert hobbies_rec[i] == {'?': 99}
                elif value in (1, 2):
                    assert hobbies_rec[i] == 1
                elif value in (3, 4):
                    assert hobbies_rec[i] == 2

        # 3.2 combine_responses.

        cat_map = {
            1: [1, 2],
            2: [97],
            3: [98, 99]
        }

        cat_names = {
            1: 'music_recoded_1',
            2: 'music_recoded_2',
            3: 'music_recoded_3'
        }
        new_var = dataset.combine_multiple_response(
            'music',
            map=cat_map,
            categories=cat_names,
            name='Music (alt)',
            alias='music_recoded'
        )
        assert isinstance(new_var, Variable)
        assert new_var.type == 'multiple_response'

        df = pandaslib.dataframe(dataset.resource)
        assert 'music_recoded' in df

        # TODO: Test the data in the recoded variable. Unsure of its meaning.

    finally:
        dataset.resource.delete()


if __name__ == '__main__':
    main()
    exit(0)



from pycrunch.importing import Importer
from scrunch.datasets import BaseDataset, _get_dataset
from scrunch.exceptions import InvalidDatasetTypeError
from scrunch.helpers import shoji_entity_wrapper


def get_streaming_dataset(dataset, connection=None, editor=False, project=None):
    """
    A simple wrapper of _get_dataset with streaming=True
    """
    shoji_ds, root = _get_dataset(dataset, connection, editor, project)
    # make sure the Dataset is of type streaming != "streaming"
    if shoji_ds['body'].get('streaming') != 'streaming':
        raise InvalidDatasetTypeError("Dataset %s is of type 'mutable',\
            use get_mutable_dataset method instead" % dataset)
    ds = StreamingDataset(shoji_ds)
    if editor is True:
        authenticated_url = root.urls["user_url"]
        ds.change_editor(authenticated_url)
    return ds


class StreamingDataset(BaseDataset):
    """
    A Crunch entity that represents Datasets that are currently
    of the "streaming" class
    """

    def stream_rows(self, columns):
        """
        Receives a dict with columns of values to add and streams them
        into the dataset. Client must call .push_rows(n) later or wait until
        Crunch automatically processes the batch.

        Returns the total of rows streamed
        """
        importer = Importer()
        count = len(list(columns.values())[0])
        for x in range(count):
            importer.stream_rows(self.resource,
                                 {a: columns[a][x] for a in columns})
        return count

    def push_rows(self, count=None):
        """
        Batches in the rows that have been recently streamed. This forces
        the rows to appear in the dataset instead of waiting for crunch
        automatic batcher process.
        """
        if bool(self.resource.stream.body.pending_messages):
            self.resource.batches.create(
                shoji_entity_wrapper({
                    'stream': count,
                    'type': 'ldjson'}
                ))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import mock
from pycrunch import ClientError, shoji, lemonpy

import scrunch
from scrunch.variables import validate_variable_url
from scrunch import get_project, get_mutable_dataset, get_user
from scrunch.datasets import Project, User
from scrunch.mutable_dataset import MutableDataset
from scrunch.order import Path
from scrunch.exceptions import InvalidPathError

from .mock_session import MockSession


@pytest.fixture(scope='function')
def envpatch():
    import os
    os.environ['CRUNCH_USERNAME'] = 'USERNAME'
    os.environ['CRUNCH_PASSWORD'] = 'PASSWORD'


def test_variable_url_validation():
    ds_url = 'https://test.crunch.io/api/datasets/b4d10b49c385aa405756fbbf572649d3/'
    assert not validate_variable_url(ds_url)

    var_url = (
        'https://test.crunch.io/api/datasets/b4d10b49c385aa405756fbbf572649d3'
        '/variables/b4d10b49c385aa405756fbbf572649d3/'
    )
    assert validate_variable_url(var_url)

    var_wo_trailing_slash = (
        'https://test.crunch.io/api/datasets/b4d10b49c385aa405756fbbf572649d3'
        '/variables/b4d10b49c385aa405756fbbf572649d3'
    )
    assert validate_variable_url(var_wo_trailing_slash)

    subvar_url = (
        'https://test.crunch.io/api/datasets/b4d10b49c385aa405756fbbf572649d3'
        '/variables/b4d10b49c385aa405756fbbf572649d3'
        '/subvariables/b4d10b49c385aa405756fbbf572649d3/'
    )
    assert validate_variable_url(subvar_url)

    subvar_wo_trailing_slash = (
        'https://test.crunch.io/api/datasets/b4d10b49c385aa405756fbbf572649d3'
        '/variables/b4d10b49c385aa405756fbbf572649d3'
        '/subvariables/b4d10b49c385aa405756fbbf572649d3'
    )
    assert validate_variable_url(subvar_wo_trailing_slash)

    var_catalog = (
        'https://test.crunch.io/api/datasets/b4d10b49c385aa405756fbbf572649d3'
        '/variables/b4d10b49c385aa405756fbbf572649d3/summary'
    )
    assert not validate_variable_url(var_catalog)


def _by_side_effect(shoji, entity_mock):
    d = {
        'id': {shoji['body']['id']: entity_mock},
        'name': {shoji['body']['name']: entity_mock},
    }

    def _get(*args):
        return d.get(args[0])
    return _get


def mocked_project_get_child(name):
    session = MockSession()
    session.feature_flags = {'old_projects_order': False}
    res_url = 'http://example.com/api/projects/A/'
    payload = {
        'element': 'shoji:entity',
        'self': res_url,
        'catalogs': {},
        'body': {'name': name, 'id': 1},
        'index': {},
        'graph': []
    }
    session.add_fixture(res_url, payload)

    return Project(session.get(res_url).payload)


class TestUtilities(object):

    @mock.patch('pycrunch.session')
    def test_get_connection_with_session(self, session_mock):
        assert scrunch.datasets._get_connection() == session_mock

    @mock.patch('scrunch.datasets.connect')
    def test_get_connection_with_ini(self, connect_mock):
        import os
        current_dir = os.path.dirname(os.path.abspath(__file__))
        ini_file = os.path.join(current_dir, 'crunch_test.ini')
        scrunch.datasets._get_connection(ini_file)
        user = 'testuser@yougov.com'
        pw = 'supersecret'
        site = 'https://test.crunch.io/api/'
        assert connect_mock.mock_calls[0].kwargs == {'site_url': site, 'username': user, 'pw': pw}

    @mock.patch('scrunch.datasets.connect')
    def test_get_connection_with_ini_api_key(self, connect_mock):
        import os
        current_dir = os.path.dirname(os.path.abspath(__file__))
        ini_file = os.path.join(current_dir, 'crunch_test_api_key.ini')
        scrunch.datasets._get_connection(ini_file)
        api_key = 'apikeysupersecret'
        site = 'https://test.crunch.io/api/'
        assert connect_mock.mock_calls[0].kwargs == {'site_url': site, 'api_key': api_key}

    @mock.patch('scrunch.datasets.connect')
    def test_get_connection_with_env(self, connect_mock, envpatch):
        import os
        user = os.environ.get('CRUNCH_USERNAME')
        pw = os.environ.get('CRUNCH_PASSWORD')
        assert user == 'USERNAME'
        assert pw == 'PASSWORD'
        scrunch.datasets._get_connection()
        assert connect_mock.mock_calls[0].kwargs == {'username': user, 'pw': pw}

    @mock.patch('pycrunch.session')
    def test_get_dataset(self, root):
        session = root.session
        dataset_url = "/api/datasets/123456/"
        shoji_entity = {
            'element': 'shoji:entity',
            'body': {
                'id': '123456',
                'name': 'dataset_name',
                'streaming': 'no'
            },
            'catalogs': {
                "variables": "variables/"
            }
        }

        def _get(url, *args, **kwargs):
            if url == '/api/datasets/dataset_name/':
                # This is the first attempt to get by ID, which should fail
                resp = mock.MagicMock(status_code=404)
                raise ClientError(resp)
            elif url == dataset_url:
                # This is when we GET by the correct URL that has been
                # discovered by using `datasets_by_name`
                payload = mock.MagicMock(**shoji_entity)
                payload.self = dataset_url
                return mock.MagicMock(payload=payload)

        session.get.side_effect = _get

        root.catalogs.datasets = '/api/datasets/'
        catalog_url = lemonpy.URL("/api/datasets/by_name/", None)

        # Calling root.follow should return the "by_name" catalog
        by_name_catalog = shoji.Catalog(session, **{
            "index": shoji.Index(session, catalog_url, **{
                dataset_url: shoji.Tuple(session, dataset_url, **{
                    "name": "dataset_name"
                })
            })
        })
        root.follow.return_value = by_name_catalog
        ds = get_mutable_dataset('dataset_name')

        # We called indeed the `datasets_by_name` with the right arg
        root.follow.assert_called_with("datasets_by_name", {
            "name": "dataset_name"
        })

        assert isinstance(ds, MutableDataset)
        assert ds.name == 'dataset_name'
        assert ds.url == dataset_url

    @mock.patch('pycrunch.session')
    def test_get_dataset_from_project_no_name(self, session):
        dataset_id = 'b2c4c6b7d3a94e58937b23c1fed1b65e'
        shoji_entity = {
            'element': 'shoji:entity',
            'body': {
                'id': dataset_id,
                'name': 'dataset_name',
                'streaming': 'no'
            }
        }

        ds_res = mock.MagicMock(**shoji_entity)
        ds_res.entity = mock.MagicMock(**shoji_entity)
        session.datasets.by.side_effect = KeyError()

        response = mock.MagicMock()
        response.payload = ds_res

        def _get(*args, **kwargs):
            return response

        session.session.get.side_effect = _get
        session.catalogs.datasets = 'https://test.crunch.io/api/'

        ds = get_mutable_dataset(dataset_id)
        session.session.get.assert_called_with('https://test.crunch.io/api/b2c4c6b7d3a94e58937b23c1fed1b65e/')

        assert isinstance(ds, MutableDataset)
        assert ds.name == 'dataset_name'
        assert ds.id == dataset_id

    @mock.patch('pycrunch.session')
    def test_get_project(self, session):

        shoji_entity = {
            "element": "shoji:catalog",
            "body": {
                "name": "Y Team",
                "id": "614a7b2ebe9a4292bba54edce83563ae"
            }
        }

        site_mock = mock.MagicMock(**shoji_entity)
        site_mock.entity = mock.MagicMock(**shoji_entity)
        session.projects.by.side_effect = _by_side_effect(shoji_entity, site_mock)

        project = get_project('Y Team')
        session.projects.by.assert_called_with('name')
        assert isinstance(project, Project)
        assert project.id == '614a7b2ebe9a4292bba54edce83563ae'

    @mock.patch('pycrunch.session')
    @mock.patch('scrunch.datasets.Project.get_child', side_effect=mocked_project_get_child)
    def test_get_project_nested(self, mocked_get_child, session):

        shoji_entity = {
            "element": "shoji:catalog",
            "body": {
                "name": "Y Team",
                "id": "614a7b2ebe9a4292bba54edce83563ae"
            },
            "index": {
                "9165e5f4eb004bb4b257a90645bfb968": {
                    "name": "California",
                    "id": "9165e5f4eb004bb4b257a90645bfb968",
                    "type": "project"
                }
            }
        }

        site_mock = mock.MagicMock(**shoji_entity)
        site_mock.entity = mock.MagicMock(**shoji_entity)
        session.projects.by.side_effect = _by_side_effect(shoji_entity, site_mock)

        project = get_project('Y Team|California')
        assert isinstance(project, Project)
        assert str(project) == 'California'
        mocked_get_child.assert_called_with('California')

    @mock.patch('pycrunch.session')
    def test_get_project_by_id(self, session):

        shoji_entity = {
            "element": "shoji:catalog",
            "body": {
                "name": "Y Team",
                "id": "614a7b2ebe9a4292bba54edce83563ae"
            }
        }

        site_mock = mock.MagicMock(**shoji_entity)
        site_mock.entity = mock.MagicMock(**shoji_entity)
        session.projects.by.side_effect = _by_side_effect(shoji_entity, site_mock)

        project = get_project('614a7b2ebe9a4292bba54edce83563ae')
        session.projects.by.assert_called_with('id')
        assert isinstance(project, Project)

        with pytest.raises(KeyError) as excinfo:
            get_project('invalidid')
        assert str(excinfo.value) == \
               "'Project (name or id: invalidid) not found.'"
        # ^ That exception message is wrapped in quotes? Ugh?

    @mock.patch('pycrunch.session')
    def test_get_user(self, session):

        shoji_entity = {
            "element": "shoji:catalog",
            "body": {
                "name": "Heisenberg",
                "email": "heisenberg@sc.org",
            }
        }

        site_mock = mock.MagicMock(**shoji_entity)
        site_mock.entity = mock.MagicMock(**shoji_entity)
        session.users.by.side_effect = {"heisenberg@sc.org": site_mock},

        user = get_user('heisenberg@sc.org')
        session.users.by.assert_called_with('email')
        assert isinstance(user, User)

    def test_path(self):
        with pytest.raises(TypeError) as excinfo:
            Path(1234)
        assert str(excinfo.value) == 'The path must be a string object'

        Path('|If Skadeförsäkring')
        Path('|æøå')

        path = Path('test')
        assert path.is_relative is True
        assert repr(path) == 'test'

    def test_user_agent(self):
        user = 'testuser@yougov.com'
        pw = 'supersecret'
        site = 'https://test.crunch.io/api/'
        from pycrunch.version import __version__ as pycrunch_v
        from requests.sessions import Session
        with mock.patch.object(Session, 'send') as mock_send:
            scrunch.connect(user, pw, site)
        prep_req = mock_send.call_args[0][0]
        assert prep_req.headers['user-agent'] == 'scrunch/%s (pycrunch/%s)' % (scrunch.__version__, pycrunch_v)



# coding: utf-8

import os
from unittest import TestCase

from scrunch import connect
from scrunch.mutable_dataset import get_mutable_dataset

HOST = os.environ['SCRUNCH_HOST']
username = os.environ['SCRUNCH_USER']
password = os.environ['SCRUNCH_PASS']


site = connect(username, password, HOST)
assert site is not None, "Unable to connect to %s" % HOST

as_entity = lambda b: {
    "element": "shoji:entity",
    "body": b
}


class TestDatasetMethods(TestCase):
    def test_replace_values(self):
        ds = site.datasets.create(as_entity({"name": "test_replace_values"})).refresh()
        variable = ds.variables.create(
            as_entity(
                {
                    "name": "my_var",
                    "alias": "my_var",
                    "type": "numeric",
                    "values": [1, 2, 3, None, 5],
                }
            )
        ).refresh()
        scrunch_dataset = get_mutable_dataset(ds.body.id, site)
        resp = scrunch_dataset.replace_values({
            "my_var": 4
        }, filter="missing(my_var)")
        if resp is not None:
            # We got a 202 response. Scrunch should have waited for the
            # progress to finish
            progress_url = resp.payload["value"]
            progress = site.session.get(progress_url)
            progress_status = progress.payload["value"]
            assert progress_status == {'progress': 100, 'message': 'complete'}
        else:
            # This means the API handled this synchronously. 204 response
            pass

        r = ds.follow("table", "limit=10")["data"]
        assert r[variable.body["id"]] == [1, 2, 3, 4, 5]
        ds.delete()


class TestCategories(TestCase):
    def test_edit_category(self):
        ds = site.datasets.create(as_entity({"name": "test_edit_category"})).refresh()

        categories = [
            {"id": 1, "name": "One", "missing": False, "numeric_value": None},
            {"id": 2, "name": "Two", "missing": False, "numeric_value": None},
            {"id": -1, "name": "No Data", "missing": True, "numeric_value": None}
        ]

        my_cat = ds.variables.create(as_entity({
            "name": "my_cat",
            "alias": "my_cat",
            "type": "categorical",
            "categories": categories
        }))

        scrunch_dataset = get_mutable_dataset(ds.body.id, site)
        my_cat = scrunch_dataset[my_cat.body["alias"]]  # Ensure refreshed var
        my_cat.categories[1].edit(numeric_value=1)

        my_cat_reloaded = scrunch_dataset[my_cat.alias]
        assert my_cat_reloaded.categories[1].as_dict() == dict(categories[0], numeric_value=1, selected=False)
        assert my_cat_reloaded.categories[2].as_dict() == dict(categories[1], selected=False)




import pytest
from mock import MagicMock
from unittest import TestCase
from scrunch.datasets import Variable
from scrunch.categories import CategoryList


def TEST_CATEGORIES():
    return [
        {"id": 1, "name": "Female", "missing": False, "numeric_value": None},
        {"id": 2, "name": "Male", "missing": False, "numeric_value": None},
        {"id": -1, "name": "No Data", "missing": True, "numeric_value": None}
    ]


def TEST_CATEGORIES_WITH_DATE():
    return [
        {"id": 1, "name": "Female", "missing": False, "numeric_value": None, "date": "2020-01-01"},
        {"id": 2, "name": "Male", "missing": False, "numeric_value": None, "date": "2020-02-02"},
        {"id": -1, "name": "No Data", "missing": True, "numeric_value": None}
    ]


class EditableMock(MagicMock):
    def edit(self, **kwargs):
        self.body.update(kwargs)
        self._edit(**kwargs)


class TestCategories(TestCase):
    def test_instance_is_reused(self):
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES(),
            type='categorical'
        )
        variable = Variable(resource, MagicMock())
        cat_list = variable.categories
        self.assertTrue(isinstance(cat_list, CategoryList))

    def test_category_dict_attribytes(self):
        resource = EditableMock()
        resource.entity.body = dict(categories=TEST_CATEGORIES(), type='categorical')
        variable = Variable(resource, MagicMock())
        # Does not have `date` unnecessarily
        assert variable.categories[1].as_dict() == {
            'id': 1,
            'missing': False,
            'name': 'Female',
            'numeric_value': None,
            'selected': False
        }
        variable.categories[1].edit(date="1990-02-04")
        # Contains .date if needed
        assert variable.categories[1].as_dict() == {
            'id': 1,
            'missing': False,
            'name': 'Female',
            'numeric_value': None,
            'selected': False,
            "date": "1990-02-04"
        }

    def test_edit_category(self):
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES(),
            type='categorical'
        )
        variable = Variable(resource, MagicMock())
        variable.categories[1].edit(name='Mujer')
        resource.entity._edit.assert_called_with(categories=[
            {'numeric_value': None, 'selected': False, 'id': 1, 'missing': False, 'name': 'Mujer'},
            # These two don't have selected yet because it is reusing the
            # API categories still, only replacing the modified one
            {'numeric_value': None, 'missing': False, 'id': 2, 'name': 'Male'},
            {'numeric_value': None, 'missing': True, 'id': -1, 'name': 'No Data'}
        ])
        resource.entity.refresh.assert_called_once()
        self.assertEqual(variable.categories[1].name, 'Mujer')

        # Editing Male
        variable.categories[2].edit(name='Hombre')
        resource.entity._edit.assert_called_with(categories=[
            {'numeric_value': None, 'selected': False, 'id': 1, 'missing': False, 'name': 'Mujer'},
            {'numeric_value': None, 'selected': False, 'missing': False, 'id': 2, 'name': 'Hombre'},
            # Same as above, reusing the existing value from API still
            {'numeric_value': None, 'missing': True, 'id': -1, 'name': 'No Data'}
        ])

        # Try to change the ID
        with self.assertRaises(AttributeError) as err:
            variable.categories[2].edit(id=100)
        self.assertEqual(
            str(err.exception),
            'Cannot edit the following attributes: id'
        )

        # Nothing changed
        self.assertEqual(set(variable.categories.keys()), {1, 2, -1})

    def test_delete_category(self):
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES(),
            type='categorical'
        )
        variable = Variable(resource, MagicMock())
        variable.categories[1].delete()
        # Calling edit without the one that we wanted to delete
        resource.entity._edit.assert_called_with(categories=[
            {'numeric_value': None, 'missing': False, 'id': 2, 'name': 'Male'},
            {'numeric_value': None, 'missing': True, 'id': -1, 'name': 'No Data'}
        ])

    def test_category_attribute_writes(self):
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES(),
            type='categorical',
        )
        variable = Variable(resource, MagicMock())

        error_msg = "use the edit() method for mutating attributes"

        with pytest.raises(AttributeError) as excinfo:
            variable.categories[1].id = 42
        # nothing has changed
        assert variable.categories[1].id == 1
        assert str(excinfo.value) == "Can't edit attibute 'id'"

        with pytest.raises(AttributeError) as excinfo:
            variable.categories[1].name = 'forbidden'
        # nothing has changed
        assert variable.categories[1].name == 'Female'
        assert str(excinfo.value) == error_msg

        with pytest.raises(AttributeError) as excinfo:
            variable.categories[1].numeric_value = 42
        # nothing has changed
        assert variable.categories[1].numeric_value is None
        assert str(excinfo.value) == error_msg

        with pytest.raises(AttributeError) as excinfo:
            variable.categories[1].missing = True
        # nothing has changed
        assert variable.categories[1].missing is False
        assert str(excinfo.value) == error_msg

        with pytest.raises(AttributeError) as excinfo:
            variable.categories[1].selected = True
        # nothing has changed, default is False
        assert variable.categories[1].selected is False
        assert str(excinfo.value) == error_msg

    def test_edit_derived(self):
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES(),
            type='categorical',
            derivation={'function': 'derivation_function'}
        )
        variable = Variable(resource, MagicMock())

        error_msg = "Cannot edit categories on derived variables. Re-derive with the appropriate expression"
        with pytest.raises(TypeError, match=error_msg):
            variable.categories[1].edit(name='Mujer')

        # Try again with an empty derivation
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES(),
            type='categorical',
            derivation={}  # Empty

        )
        variable = Variable(resource, MagicMock())
        variable.categories[1].edit(name='Mujer')
        resource.entity._edit.assert_called_with(categories=[
            {'numeric_value': None, 'selected': False, 'id': 1, 'missing': False, 'name': 'Mujer'},
            {'numeric_value': None, 'missing': False, 'id': 2, 'name': 'Male'},
            {'numeric_value': None, 'missing': True, 'id': -1, 'name': 'No Data'}
        ])

    def test_read_category_date(self):
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES_WITH_DATE(),
            type='categorical'
        )
        variable = Variable(resource, MagicMock())
        self.assertEqual(variable.categories[1].date, '2020-01-01')
        self.assertEqual(variable.categories[2].date, '2020-02-02')
        with self.assertRaises(KeyError):
            # The `No Data` category doesn't provide a `date field
            _ = variable.categories[3].date

    def test_edit_category_date(self):
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES_WITH_DATE(),
            type='categorical'
        )
        variable = Variable(resource, MagicMock())
        variable.categories[1].edit(date='2021-01-01')
        resource.entity._edit.assert_called_with(categories=[
            {'numeric_value': None, 'selected': False, 'id': 1, 'missing': False, 'name': 'Female', 'date': '2021-01-01'},
            {'numeric_value': None, 'missing': False, 'id': 2, 'name': 'Male', 'date': '2020-02-02'},
            {'numeric_value': None, 'missing': True, 'id': -1, 'name': 'No Data'}
        ])
        resource.entity.refresh.assert_called_once()
        self.assertEqual(variable.categories[1].date, '2021-01-01')


class TestCategoryList(TestCase):
    def test_reorder(self):
        resource = EditableMock()
        resource.entity.body = dict(
            categories=TEST_CATEGORIES(),
            type='categorical'
        )
        variable = Variable(resource, MagicMock())
        variable.categories.order(2, -1, 1)

        # Reordered values
        resource.entity._edit.assert_called_with(categories=[
            {'numeric_value': None, 'missing': False, 'id': 2, 'name': 'Male'},
            {'numeric_value': None, 'missing': True, 'id': -1, 'name': 'No Data'},
            {'numeric_value': None, 'missing': False, 'id': 1, 'name': 'Female'}
        ])
        resource.entity.refresh.assert_called_once()



# coding: utf-8

"""
This example shows a more complex example using different recodes and combines
to create new variables.
"""

import csv
import os
import tempfile
from six import StringIO
from unittest import TestCase

from fixtures import NEWS_DATASET, NEWS_DATASET_ROWS, mr_in, RECODES_CSV_OUTPUT
from scrunch import connect
from scrunch.streaming_dataset import get_streaming_dataset
from scrunch.mutable_dataset import get_mutable_dataset
from pycrunch.importing import Importer


HOST = os.environ['SCRUNCH_HOST']
username = os.environ['SCRUNCH_USER']
password = os.environ['SCRUNCH_PASS']


site = connect(username, password, HOST)
assert site is not None, "Unable to connect to %s" % HOST


class TestRecodes(TestCase):
    def test_recodes(self):
        raise self.skipTest("Temporarily disabling for API update")
        # Create a dataset for usage
        ds = site.datasets.create({
            'element': 'shoji:entity',
            'body': {
                'name': 'test_recodes',
                'table': {
                    'element': 'crunch:table',
                    'metadata': NEWS_DATASET
                },
                'streaming': 'streaming'
            }
        }).refresh()
        dataset = get_streaming_dataset(ds.body.id, site)
        print("Dataset %s created" % dataset.id)

        # Add data rows
        total = dataset.stream_rows(NEWS_DATASET_ROWS)
        dataset.push_rows(total)

        # Recode a new single response variable
        agerange = dataset.create_categorical([
            {'id': 1, 'name': 'Underage', 'case': 'age < 18'},
            {'id': 2, 'name': 'Millennials', 'case': 'age > 18 and age < 25'},
            {'id': 3, 'name': 'Gen X', 'case': 'age < 35 and age >= 25'},
            {'id': 4, 'name': 'Grown ups', 'case': 'age < 60 and age >= 35'},
            {'id': 5, 'name': '60+', 'case': 'age >= 60'}
        ], alias='agerange', name='Age range', multiple=False)

        print("Variable %s created" % agerange.alias)

        # Recode a new multiple response variable from an existing multiple
        # response variable
        origintype = dataset.create_categorical([
            {'id': 1, 'name': "Online",
             # Mixed support for using "category"(subvariables really) IDs
             'case': mr_in(dataset, 'newssource', [1, 2, 3, 4])},  # Only in the helper
            {'id': 2, 'name': "Print", 'case': mr_in(dataset, 'newssource', [5, 6])},
            {'id': 3, 'name': "Tv", 'case': mr_in(dataset, 'newssource', [7, 9])},
            {'id': 4, 'name': "Radio", 'case': mr_in(dataset, 'newssource', [8, 10])},
        ], alias='origintype', name="News source by type", multiple=True)

        print("Variable %s created" % origintype.alias)

        # Add an exclusion filter
        dataset.exclude('agerange == 1')  # Remove underage

        # MAYBE: Create a new numeric expression based on arbitrary rules

        # Copy a variable
        origintype_copy = dataset.copy_variable(
            origintype,
            name='Copy of origintype',
            alias='origintype_copy')
        print("Variable %s created" % origintype_copy.alias)

        onlinenewssource = dataset.combine_categories(
            origintype_copy,
            map={
                1: 1,
                2: [2, 3, 4]
            },
            categories={
                1: "online",
                2: "notonline",
            },
            alias="onlinenewssource",
            name='Online or not')
        print('Created combination: %s' % onlinenewssource.alias)

        # Combine a single categorical - Combine with subvar 3 on the wrong place
        over35 = dataset.combine_categories(
            agerange,
            map={
                1: [1, 2, 3],
                2: [4, 5]
            },
            categories={
                1: 'under35',
                2: 'over35'
            },
            name='over 35?',
            alias='over35')
        print('Created combination: %s' % over35.alias)

        self.maxDiff = None
        # Export some rows
        output = tempfile.NamedTemporaryFile('rw', delete=True)
        dataset.export(output.name)

        reader = csv.DictReader(output)

        # put the data into columns
        actual = {}
        for row in reader:
            for k, v in row.items():
                actual.setdefault(k, []).append(v)

        reader = csv.reader(StringIO(RECODES_CSV_OUTPUT))
        headers = reader.next()

        # rewrite the actual csv in the same order as the expected csv
        actualf = StringIO()
        writer = csv.writer(actualf)
        writer.writerow(headers)
        for i in range(len(actual.values()[0])):
            writer.writerow([actual[h][i] for h in headers])
        actualf.seek(0)

        result = [l.strip() for l in actualf.read().strip().split('\n')]
        expected = RECODES_CSV_OUTPUT.split('\n')
        # Rows are unordered under streaming conditions
        self.assertEqual(sorted(result), sorted(expected))

        output.close()
        ds.delete()


class TestFill(TestCase):
    def prepare_ds(self):
        cats = [
            {"id": 1, "name": "Daily", "missing": False, "numeric_value": None},
            {"id": 2, "name": "Weekly", "missing": False,
             "numeric_value": None},
            {"id": 3, "name": "Monthly", "missing": False,
             "numeric_value": None},
            {"id": -1, "name": "No Data", "missing": True,
             "numeric_value": None},
        ]
        metadata = {
            "coke_freq": {
                "name": "frequency coke",
                "type": "categorical",
                "categories": cats
            },
            "pepsi_freq": {
                "name": "frequency pepsi",
                "type": "categorical",
                "categories": cats
            },
            "pop_pref": {
                "name": "Soda preference",
                "type": "categorical",
                "categories": [
                    {"id": 1, "name": "Coke", "missing": False,
                     "numeric_value": None},
                    {"id": 2, "name": "Pepsi", "missing": False,
                     "numeric_value": None},
                    {"id": -1, "name": "No Data", "missing": True,
                     "numeric_value": None},
                ]
            }
        }
        ds_payload = {
            'element': 'shoji:entity',
            'body': {
                'name': 'test_fill',
                'table': {
                    'element': 'crunch:table',
                    'metadata': metadata
                },
            }
        }

        rows = [
            ["coke_freq", "pepsi_freq", "pop_pref"],
            [1, 3, 1],
            [2, 2, 1],
            [3, 1, 1],
            [1, 3, 2],
            [2, 2, 2],
            [3, 1, 2],
        ]
        ds = site.datasets.create(ds_payload).refresh()
        dataset = get_mutable_dataset(ds.body.id, site)
        Importer().append_rows(ds, rows)
        return dataset, ds

    def test_fill(self):
        dataset, api_ds = self.prepare_ds()
        dataset.create_fill_values([
            {"case": "pop_pref == 1", "variable": "coke_freq"},
            {"case": "pop_pref == 2", "variable": "pepsi_freq"},
        ], alias="pop_freq", name="Pop frequency")

        variables = api_ds.variables.by("alias")
        new_id = variables["pop_freq"]["id"]
        new_var = variables["pop_freq"].entity
        self.assertTrue(new_var.body.derived)
        self.assertEqual(new_var.body.name, "Pop frequency")

        data = api_ds.follow("table", "limit=6")
        cats = {c["name"]: c["id"] for c in data["metadata"][new_id]["categories"]}
        self.assertEqual(data["data"][new_id], [
            # Coke chunk
            cats["Daily"],
            cats["Weekly"],
            cats["Monthly"],
            # Pepsi chunk
            cats["Monthly"],
            cats["Weekly"],
            cats["Daily"],
        ])
        api_ds.delete()

    def test_fill_w_else(self):
        dataset, ds = self.prepare_ds()
        dataset.create_fill_values([
            {"case": "pop_pref == 1", "variable": "coke_freq"},
            {"case": "else", "variable": "pepsi_freq"},
        ], alias="pop_freq", name="Pop frequency")

        variables = ds.variables.by("alias")
        new_id = variables["pop_freq"]["id"]
        new_var = variables["pop_freq"].entity
        self.assertTrue(new_var.body.derived)
        self.assertEqual(new_var.body.name, "Pop frequency")

        data = ds.follow("table", "limit=6")
        cats = {c["name"]: c["id"] for c in data["metadata"][new_id]["categories"]}
        self.assertEqual(data["data"][new_id], [
            # Coke chunk
            cats["Daily"],
            cats["Weekly"],
            cats["Monthly"],
            # Pepsi chunk - Default case
            cats["Monthly"],
            cats["Weekly"],
            cats["Daily"],
        ])
        ds.delete()

    def test_fill_w_else_code(self):
        dataset, ds = self.prepare_ds()
        dataset.create_fill_values([
            {"case": "pop_pref == 1", "variable": "coke_freq"},
            {"case": "else", "name": "Not Asked", "id": 99, "missing": False},
        ], alias="pop_freq", name="Pop frequency")

        variables = ds.variables.by("alias")
        new_id = variables["pop_freq"]["id"]
        new_var = variables["pop_freq"].entity
        self.assertTrue(new_var.body.derived)
        self.assertEqual(new_var.body.name, "Pop frequency")

        data = ds.follow("table", "limit=6")
        cats = {c["name"]: c["id"] for c in data["metadata"][new_id]["categories"]}
        self.assertEqual(data["data"][new_id], [
            # Coke chunk
            cats["Daily"],
            cats["Weekly"],
            cats["Monthly"],
            # Default value
            cats["Not Asked"],
            cats["Not Asked"],
            cats["Not Asked"],
        ])
        ds.delete()

    def test_fill_w_else_default(self):
        dataset, ds = self.prepare_ds()
        dataset.create_fill_values([
            {"case": "pop_pref == 1", "variable": "coke_freq"},
        ], alias="pop_freq", name="Pop frequency")

        variables = ds.variables.by("alias")
        new_id = variables["pop_freq"]["id"]
        new_var = variables["pop_freq"].entity
        self.assertTrue(new_var.body.derived)
        self.assertEqual(new_var.body.name, "Pop frequency")

        data = ds.follow("table", "limit=6")
        cats = {c["name"]: c["id"] for c in data["metadata"][new_id]["categories"]}
        self.assertEqual(data["data"][new_id], [
            # Coke chunk
            cats["Daily"],
            cats["Weekly"],
            cats["Monthly"],
            # Default value
            {"?": -1},
            {"?": -1},
            {"?": -1},
        ])
        ds.delete()



# coding: utf-8
from pycrunch.shoji import Catalog
from six import string_types

from scrunch.exceptions import InvalidPathError


class Folder(object):
    def __init__(self, folder_ent, root, parent):
        self.root = root
        self.parent = parent
        self.folder_ent = folder_ent
        self.name = folder_ent.body.name
        self.alias = self.name  # For compatibility with variables .alias
        self.url = folder_ent.self

    def __repr__(self):
        return '<Folder: %s>' % self.name

    def get(self, path):
        self.folder_ent.refresh()  # Always up to date
        from scrunch.order import Path

        node = self
        for p_name in Path(path).get_parts():
            try:
                node = node.get_child(p_name)
            except KeyError:
                raise InvalidPathError('Subfolder not found %s' % p)
        return node

    def __getitem__(self, path):
        return self.get(path)

    def __iter__(self):
        return self.itervalues()

    def iterkeys(self):
        for child in self.children:
            yield child.alias

    def itervalues(self):
        for child in self.children:
            yield child

    def iteritems(self):
        for child in self.children:
            yield (child.alias, child)

    def values(self):
        return list(self.itervalues())

    def keys(self):
        return list(self.iterkeys())

    def items(self):
        return list(self.iteritems())

    def get_child(self, name):
        by_name = self.folder_ent.by('name')
        by_alias = self.folder_ent.by('alias')

        # If found by alias, then it's a variable, return the variable
        if name in by_alias:
            return self.root.dataset[name]
        elif name in by_name:
            # Found by name, if it's not a folder, return the variable
            tup = by_name[name]
            if tup.type != 'folder':
                return self.root.dataset[name]
            return Folder(tup.entity, self.root, self)

        # Not a folder nor a variable
        path = self.path_pieces() + [name]
        raise InvalidPathError('Invalid path: | %s' % ' | '.join(path))

    def path_pieces(self):
        if self.parent:
            return self.parent.path_pieces() + [self.name]
        return []

    @property
    def path(self):
        return '| ' + ' | '.join(self.path_pieces())

    def create_folder(self, folder_name, position=None, after=None, before=None, alias=None):
        new_ent = self.folder_ent.create(Catalog(self.folder_ent.session, body={
            'name': folder_name
        }))
        new_ent.refresh()
        subfolder = Folder(new_ent, self.root, self)
        if position is not None or after is not None or before is not None:
            children = self._position_items([subfolder], position, before, after)
            self.reorder(children)

        if alias:
            subfolder.move_here(alias)

        self.folder_ent.refresh()
        return subfolder

    def _position_items(self, new_items, position, before, after):
        if before is not None or after is not None:
            # Before and After are strings
            target = before or after
            position = [x for x, c in enumerate(self.children) if c.alias == target]
            if not position:
                raise InvalidPathError("No child with name %s found" % target)
            position = position[0]
            if before is not None:
                position = position if position > 0 else 0
            else:
                max_pos = len(self.folder_ent.graph)
                position = (position + 1) if position < max_pos else max_pos

        if position is not None:
            new_urls = {c.url for c in new_items}
            children = [c for c in self.children if c.url not in new_urls]
            for item in reversed(new_items):
                children.insert(position, item)
            return children
        return self.children  # Nothing happened

    @property
    def children(self):
        self.folder_ent.refresh()  # Always get a fresh copy
        index = self.folder_ent.index
        ds = self.root.dataset
        _children = []
        for item_url in self.folder_ent.graph:
            if item_url not in index:
                continue
            item_tup = index[item_url]
            if item_tup['type'] == 'folder':
                subfolder = Folder(item_tup.entity, self.root, self)
                _children.append(subfolder)
            else:
                # Add the variable
                _children.append(ds[item_tup['alias']])
        return _children

    def move_here(self, *children, **kwargs):
        if not children:
            return
        children = children[0] if isinstance(children[0], list) else children
        children = [
            self.root.dataset[c] if isinstance(c, string_types) else c
            for c in children
        ]
        index = {c.url: {} for c in children}
        position, before, after = [kwargs.get('position'),
                                   kwargs.get('before'), kwargs.get('after')]
        if position is not None or after is not None or before is not None:
            children = self._position_items(children, position, before, after)
        graph = self.folder_ent.graph + [c.url for c in children]
        self.folder_ent.patch({
            'element': 'shoji:catalog',
            'index': index,
            'graph': graph
        })
        self.folder_ent.refresh()

    def append(self, *children):
        """Alias of .move_here, adds items at the end of the folder"""
        self.move_here(*children)

    def insert(self, *children, **kwargs):
        """Alias of move_here with a specific position"""
        self.move_here(*children, position=kwargs.get('position', 0))

    def move(self, path, position=None, after=None, before=None):
        """
        The inverse of move here, but it allows to move only folder by folder.
        This API is much more chatty than using .move_here() that allows to
        move multiple items in one call. Implemented for orders backwards compat
        """
        target = self.root.get(path)
        target.move_here(self, position=position, after=after, before=before)

    def rename(self, new_name):
        self.folder_ent.patch({
            'element': 'shoji:catalog',
            'body': {'name': new_name}
        })
        self.name = new_name

    def delete(self):
        self.folder_ent.delete()

    def reorder(self, *items):
        items = items[0] if isinstance(items[0], list) else items
        name2tup = self.folder_ent.by('name')
        graph = [
            name2tup[c].entity_url if isinstance(c, string_types) else c.url
            for c in items
        ]
        self.folder_ent.patch({
            'element': 'shoji:catalog',
            'graph': graph
        })
        self.folder_ent.refresh()


class DatasetFolders(object):
    def __init__(self, dataset):
        self.dataset = dataset
        folders_root = dataset.resource.folders

        # Due the rolling nature of the API we want Scrunch to work with both
        # the new Public root and old style folders API as well
        if "public" in folders_root.catalogs:
            # New style with public folders
            self.public = Folder(folders_root.public, self, None)
            self.root = Folder(folders_root.public, self, None)  # Legacy
        else:
            # Old without public root
            self.public = Folder(folders_root, self, None)
            self.root = Folder(folders_root, self, None)  # Legacy

        # Dataset viewers may not get all folders exposed
        if "hidden" in folders_root.catalogs:
            self.hidden = Folder(folders_root.hidden, self, None)
        if "secure" in folders_root.catalogs:
            self.secure = Folder(folders_root.secure, self, None)

    def get(self, path):
        return self.public.get(path)

    def __getitem__(self, path):
        return self.public.get(path)

    def __iter__(self):
        return self.public.itervalues()

    def iterkeys(self):
        for child in self.public.children:
            yield child.alias

    def itervalues(self):
        for child in self.public.children:
            yield child

    def iteritems(self):
        for child in self.public.children:
            yield (child.alias, child)

    def values(self):
        return list(self.itervalues())

    def keys(self):
        return list(self.iterkeys())

    def items(self):
        return list(self.iteritems())



# coding: utf-8

import json
import pycrunch
from pycrunch.shoji import TaskError


class ScriptExecutionError(Exception):
    def __init__(self, client_error, resolutions):
        self.client_error = client_error
        self.resolutions = resolutions

    def __repr__(self):
        return json.dumps(self.resolutions, indent=2)


class DatasetScripts:
    def __init__(self, dataset_resource):
        """
        :param dataset_resource: Pycrunch Entity for the dataset.
        """
        self.dataset_resource = dataset_resource

    def execute(self, script_body):
        try:
            self.dataset_resource.scripts.create({
                'element': 'shoji:entity',
                'body': {"body": script_body},
            })
        except pycrunch.ClientError as err:
            if isinstance(err, TaskError):
                # For async script validation error
                resolutions = err.message["resolutions"]
                raise ScriptExecutionError(err, resolutions)
            elif err.status_code == 400:
                # For sync validation
                resolutions = err.args[2]["resolutions"]
                raise ScriptExecutionError(err, resolutions)
            raise err  # 404 or something else

    def collapse(self):
        """
        When a dataset has too many scripts. Collapse will concatenate
        all the previously executed scripts into one the first. It will delete
        all savepoints associated with the collapsed scripts.
        """
        self.dataset_resource.scripts.collapse.post({})

    def all(self):
        scripts_index = self.dataset_resource.scripts.index
        scripts = []
        for s_url, s in scripts_index.items():
            scripts.append(s.entity)
        scripts = sorted(scripts, key=lambda s: s.body["creation_time"])
        return scripts

    def revert_to(self, id=None, script_number=None):
        all_scripts = self.all()
        if script_number is not None:
            script = all_scripts[script_number]
        elif id is not None:
            # We have to do this because currently the API does not expose the
            # script ID directly.
            script = [s for s in all_scripts if "scripts/{}/".format(id) in s.self][0]
        else:
            raise ValueError("Must indicate either ID or script number")

        resp = script.revert.post({})  # Asynchronous request
        pycrunch.shoji.wait_progress(resp, self.dataset_resource.session)
